﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using System.Collections.ObjectModel;

using Infragistics.Controls.Charts;
using System.Windows;
using Infragistics;
using System.Windows.Shapes;
using System.Windows.Media;
using System.Windows.Controls;


namespace ScatterLineChart
{
    public class BoxAndWhiskerData : ObservableCollection<BoxAndWhiskerDataPoint>
    {
    }

    public class BoxAndWhiskerDataPoint : DataPoint
    {
        public BoxAndWhiskerDataPoint ()
        {
        }

        public BoxAndWhiskerDataPoint (string Name, double X, double Y, double Minimum, double Maximum, double Mean, double Median, double UpperQuartile,  double LowerQuartile, double Value )
        {
            
            _dataPointName = Name;
            this.X = X;
            this.Y = Y;
            this.Minimum = Minimum;
            this.Maximum = Maximum;
            this.Mean = Mean;
            this.Median = Median;
            this.UpperQuartile = UpperQuartile;
            this.LowerQuartile = LowerQuartile;
            this.Value = Value;
        }
        //private double _y;

        //public double Y
        //{
        //    get 
        //    { 
        //        return _y; 
        //    }
        //    set
        //    {
        //        if (_y == value) return;
        //        _y = value;
        //        OnPropertyChanged("Y");
        //    }
        //}

        //private double _x;

        //public double X
        //{
        //    get 
        //    { 
        //        return _x; 
        //    }
        //    set
        //    {
        //        if (_x == value) return;
        //        _x = value;
        //        OnPropertyChanged("X");
        //    }
        //}
        

        private double _value;

        public double Value
        {
            get
            {
                return _value;
            }
            set
            {
                if (value != _value)
                {
                    _value = value;
                    OnPropertyChanged("Value");
                }
            }
        }

        private bool _isEmpty = false;

        public bool IsEmpty
        {
            get
            {
                return _isEmpty;
            }
            set
            {
                _isEmpty = value;
            }
        }

        private string _dataPointName;

        public string DataPointName
        {
            get
            {
                return _dataPointName;
            }
            set
            {
                if (value != _dataPointName)
                {
                    _dataPointName = value;
                    OnPropertyChanged("DataPointName");
                }
            }
        }

        private double _mean;

        public double Mean
        {
            get
            {
                return _mean;
            }
            set
            {
                if (value != _mean)
                {
                    _mean = value;
                    OnPropertyChanged("Mean");
                }
            }
        }

        private double _median;

        public double Median
        {
            get
            {
                return _median;
            }
            set
            {
                if (value != _median)
                {
                    _median = value;
                    OnPropertyChanged("Median");
                }
            }
        }

        private double _upperQuartile;

        public double UpperQuartile
        {
            get
            {
                return _upperQuartile;
            }
            set
            {
                if (value != _upperQuartile)
                {
                    _upperQuartile = value;
                    OnPropertyChanged("UpperQuartile");
                }
            }
        }

        private double _lowerQuartile;

        public double LowerQuartile
        {
            get
            {
                return _lowerQuartile;
            }
            set
            {
                if (value != _lowerQuartile)
                {
                    _lowerQuartile = value;
                    OnPropertyChanged("LowerQuartile");
                }
            }
        }

        private double _minimum;

        public double Minimum
        {
            get
            {
                return _minimum;
            }
            set
            {
                if (value != _minimum)
                {
                    _minimum = value;
                    OnPropertyChanged("Minimum");
                }
            }
        }

        private double _maximum;

        public double Maximum
        {
            get
            {
                return _maximum;
            }
            set
            {
                if (value != _maximum)
                {
                    _maximum = value;
                    OnPropertyChanged("Maximum");
                }
            }
        }

        public string FormattedLabel
        {
            get
            {
                if (!IsEmpty)
                {
                    return String.Format("Max:  {0:F3} \nMin:  {1:F3} \nUQ:   {2:F3} \nLQ:   {3:F3} \nMed:  {4:F3} \nMean: {5:F3}", Maximum, Minimum, UpperQuartile, LowerQuartile, Median, Mean);
                    //return Label;
                }

                return null;
            }
        }

        public new string ToString()
        {
            return String.Format("Value {0}, X {1}, Y {2}", Value, X, Y);
        }
    }

    public class BoxAndWhiskerSeries : HorizontalAnchoredCategorySeries
    {
        public BoxAndWhiskerSeries()
        {
            this.DefaultStyleKey = typeof(BoxAndWhiskerSeries);
        }

        protected override void RenderThumbnail(Rect viewportRect, RenderSurface surface)
        {
            base.RenderThumbnail(viewportRect, surface);

            // check if the series can be rendered:
            // - the Viewport (the bounds rectangle for the series) is not empty, 
            // - the RootCanvas (the container for the custom graphics) is not null.  
            // - the Axes are not null.  
            // - the ItemsSource is not null.  
            if (this.Viewport.IsEmpty || this.RootCanvas == null ||
                this.XAxis == null || this.YAxis == null ||
                this.ItemsSource == null || surface == null || surface.Surface == null)
            {
                return;
            }

            // clears the RootCanvas on every render of the series
            surface.Surface.Children.Clear();

            Rect surfaceRect = new Rect(0.0, 0.0, 1.0, 1.0);

            // create data structure for boxAndWhiskers based on values of items in the source of this series
            BoxAndWhiskerData data = (BoxAndWhiskerData)this.ItemsSource;

            int index = 0;
            int contourIndex = 0;

            // re-use chart's brushes and outlines for actual boxAndWhisker's brushes and outlines 
            this.ActualBoxAndWhiskerBrushes = ((XamDataChart)this.SeriesViewer).Brushes;
            this.ActualBoxAndWhiskerOutlines = ((XamDataChart)this.SeriesViewer).MarkerOutlines;

            double scale = viewportRect.Width / this.Viewport.Width;

            foreach (BoxAndWhiskerDataPoint dataPoint in data)
            {
                if (dataPoint.IsEmpty)
                {
                    index++;
                    continue;
                }

                DrawBoxAndWhisker(surfaceRect, viewportRect, surface.Surface, dataPoint, index, contourIndex, scale);

                index++;
            }
        }

        private void DrawBoxAndWhisker(Rect surfaceRect, Rect viewportRect, Canvas drawingCanvas, BoxAndWhiskerDataPoint dataPoint, int index, int contourIndex, double scale)
        {
            bool isInverted = this.XAxis.IsInverted;

            ScalerParams param = new ScalerParams(surfaceRect, viewportRect, isInverted);

            // scale x-locations of data point to the series' viewport
            double x = this.XAxis.GetScaledValue(index++, param);

            isInverted = this.YAxis.IsInverted;
            param = new ScalerParams(surfaceRect, viewportRect, isInverted);

            // scale y-location of data point to the series' viewport
            double top = this.YAxis.GetScaledValue(dataPoint.UpperQuartile, param);
            double max = this.YAxis.GetScaledValue(dataPoint.Maximum, param);
            double min = this.YAxis.GetScaledValue(dataPoint.Minimum, param);
            double mean = this.YAxis.GetScaledValue(dataPoint.Mean, param);
            double median = this.YAxis.GetScaledValue(dataPoint.Median, param);
            double bottom = this.YAxis.GetScaledValue(dataPoint.LowerQuartile, param);

            double boxAndWhiskerWhiskerWidth = 50 * scale;

            // get parameters of a boxAndWhisker marker
            double boxAndWhiskerMarkerValue = dataPoint.Value;
            double boxAndWhiskerMarkerWidth = 100 * scale;
            double boxAndWhiskerMarkerHeight = bottom - top;

            double boxAndWhiskerMarkerLocationLeft = x - boxAndWhiskerMarkerWidth / 2;
            double boxAndWhiskerMarkerLocationTop = top;

            // create element for shape of a boxAndWhisker marker
            Rectangle boxAndWhiskerMarker = new Rectangle();
            boxAndWhiskerMarker.Fill = GetBoxAndWhiskerMarkerFill(contourIndex);

            boxAndWhiskerMarker.Stroke = GetBoxAndWhiskerMarkerOutline(contourIndex);
            boxAndWhiskerMarker.StrokeThickness = 1.0;
            boxAndWhiskerMarker.Width = System.Math.Max(0.0, boxAndWhiskerMarkerWidth);
            boxAndWhiskerMarker.Height = System.Math.Max(0.0, boxAndWhiskerMarkerHeight);

            // create element to hold elements of a boxAndWhisker marker 
            Grid markerGrid = new Grid();
            markerGrid.Children.Add(boxAndWhiskerMarker);
            Canvas.SetLeft(markerGrid, boxAndWhiskerMarkerLocationLeft);
            Canvas.SetTop(markerGrid, boxAndWhiskerMarkerLocationTop);
            Canvas.SetZIndex(markerGrid, contourIndex + 11);

            // render the marker of the current boxAndWhisker on the canvas of this series
            drawingCanvas.Children.Add(markerGrid);

            // Render whiskers
            double strokeThickness = 2.0;

            Line maximumLine = new Line();
            maximumLine.Stroke = new SolidColorBrush(Color.FromRgb(0, 0, 0));
            maximumLine.StrokeThickness = strokeThickness;
            maximumLine.X1 = x;
            maximumLine.X2 = x;
            maximumLine.Y1 = top;
            maximumLine.Y2 = max;

            drawingCanvas.Children.Add(maximumLine);

            Line maximumCrossLine = new Line();
            maximumCrossLine.Stroke = new SolidColorBrush(Color.FromRgb(0, 0, 0));
            maximumCrossLine.StrokeThickness = strokeThickness;
            maximumCrossLine.X1 = x - boxAndWhiskerWhiskerWidth / 2;
            maximumCrossLine.X2 = x + boxAndWhiskerWhiskerWidth / 2;
            maximumCrossLine.Y1 = max;
            maximumCrossLine.Y2 = max;

            drawingCanvas.Children.Add(maximumCrossLine);

            Line minimumLine = new Line();
            minimumLine.Stroke = new SolidColorBrush(Color.FromRgb(0, 0, 0));
            minimumLine.StrokeThickness = strokeThickness;
            minimumLine.X1 = x;
            minimumLine.X2 = x;
            minimumLine.Y1 = bottom;
            minimumLine.Y2 = min;

            drawingCanvas.Children.Add(minimumLine);

            Line minimumCrossLine = new Line();
            minimumCrossLine.Stroke = new SolidColorBrush(Color.FromRgb(0, 0, 0));
            minimumCrossLine.StrokeThickness = strokeThickness;
            minimumCrossLine.X1 = x - boxAndWhiskerWhiskerWidth / 2;
            minimumCrossLine.X2 = x + boxAndWhiskerWhiskerWidth / 2;
            minimumCrossLine.Y1 = min;
            minimumCrossLine.Y2 = min;

            drawingCanvas.Children.Add(minimumCrossLine);

            // Render median
            Line medianLine = new Line();
            medianLine.Stroke = new SolidColorBrush(Color.FromRgb(50, 50, 50));
            medianLine.StrokeThickness = strokeThickness;
            medianLine.X1 = x - boxAndWhiskerMarkerWidth / 2;
            medianLine.X2 = x + boxAndWhiskerMarkerWidth / 2;
            medianLine.Y1 = median;
            medianLine.Y2 = median;
            Canvas.SetZIndex(medianLine, contourIndex + 15);

            drawingCanvas.Children.Add(medianLine);

            // Render mean
            double meanSize = 20 * scale;

            Ellipse meanEllipse = new Ellipse();
            meanEllipse.Stroke = new SolidColorBrush(Color.FromRgb(50, 50, 50));
            meanEllipse.Fill = new SolidColorBrush(Color.FromRgb(50, 50, 50));
            meanEllipse.StrokeThickness = strokeThickness;
            meanEllipse.Width = meanSize;
            meanEllipse.Height = meanSize;

            Canvas.SetLeft(meanEllipse, x - meanSize / 2.0);
            Canvas.SetTop(meanEllipse, mean - meanSize / 2.0);
            Canvas.SetZIndex(meanEllipse, contourIndex + 20);

            drawingCanvas.Children.Add(meanEllipse);
        }

        protected override void RenderSeriesOverride(bool animate)
        {
            base.RenderSeriesOverride(animate);

            // check if the series can be rendered:
            // - the Viewport (the bounds rectangle for the series) is not empty, 
            // - the RootCanvas (the container for the custom graphics) is not null.  
            // - the Axes are not null.  
            // - the ItemsSource is not null.  
            if (this.Viewport.IsEmpty || this.RootCanvas == null ||
                this.XAxis == null || this.YAxis == null ||
                this.ItemsSource == null)
            {
                return;
            }

            // clears the RootCanvas on every render of the series
            this.RootCanvas.Children.Clear();

            // create data structure for boxAndWhiskers based on values of items in the source of this series
            BoxAndWhiskerData data = (BoxAndWhiskerData)this.ItemsSource;

            int index = 0;
            int contourIndex = 0;

            // re-use chart's brushes and outlines for actual boxAndWhisker's brushes and outlines 
            this.ActualBoxAndWhiskerBrushes = ((XamDataChart)this.SeriesViewer).Brushes;
            this.ActualBoxAndWhiskerOutlines = ((XamDataChart)this.SeriesViewer).MarkerOutlines;

            double scale = 1.0;

            foreach (BoxAndWhiskerDataPoint dataPoint in data)
            {
                if (dataPoint.IsEmpty)
                {
                    index++;
                    continue;
                }

                DrawBoxAndWhisker(this.SeriesViewer.WindowRect, this.Viewport, this.RootCanvas, dataPoint, index, contourIndex, scale);

                index++;
            }
        }

        public BrushCollection ActualBoxAndWhiskerBrushes { get; private set; }
        public BrushCollection ActualBoxAndWhiskerOutlines { get; private set; }

        private Brush GetBoxAndWhiskerPathFill(int contourIndex)
        {
            return GetValidBrush(contourIndex, ((XamDataChart)this.SeriesViewer).Brushes);
        }
        private Brush GetBoxAndWhiskerPathStroke(int contourIndex)
        {
            return GetValidBrush(contourIndex, ((XamDataChart)this.SeriesViewer).MarkerOutlines);
        }
        private Brush GetBoxAndWhiskerMarkerOutline(int contourIndex)
        {
            return GetValidBrush(contourIndex, ((XamDataChart)this.SeriesViewer).MarkerOutlines);
        }
        private Brush GetBoxAndWhiskerMarkerFill(int contourIndex)
        {
            return GetValidBrush(contourIndex, ((XamDataChart)this.SeriesViewer).MarkerBrushes);
        }
        private Brush GetValidBrush(int contourIndex, BrushCollection brushes)
        {
            if (brushes == null || brushes.Count == 0)
            {
                return new SolidColorBrush(Colors.Black);
            }
            if (contourIndex >= 0 && contourIndex < brushes.Count)
            {
                return brushes[contourIndex];
            }
            contourIndex = contourIndex % brushes.Count;
            return brushes[contourIndex];
        }

        public const string XAxisPropertyName = "XAxis";

        public const string YAxisPropertyName = "YAxis";

        /// <summary>
        /// Redraw series any time the Viewport rect changes
        /// </summary>
        /// <param name="oldViewportRect"></param>
        /// <param name="newViewportRect"></param>
        protected override void ViewportRectChangedOverride(Rect oldViewportRect, Rect newViewportRect)
        {
            base.ViewportRectChangedOverride(oldViewportRect, newViewportRect);
            this.RenderSeries(false);
        }
        protected override void WindowRectChangedOverride(Rect oldWindowRect, Rect newWindowRect)
        {
            base.WindowRectChangedOverride(oldWindowRect, newWindowRect);
            this.RenderSeries(false);
        }

        /// <summary>
        /// Overrides property update of this series
        /// </summary>
        protected override void PropertyUpdatedOverride(object sender, string propertyName, object oldValue, object newValue)
        {
            base.PropertyUpdatedOverride(sender, propertyName, oldValue, newValue);
            switch (propertyName)
            {
                case ItemsSourcePropertyName:
                    this.RenderSeries(false);
                    if (this.XAxis != null)
                    {
                        this.XAxis.UpdateRange();
                    }
                    if (this.YAxis != null)
                    {
                        this.YAxis.UpdateRange();
                    }
                    break;

                case XAxisPropertyName:
                    if (oldValue != null)
                    {
                        ((Axis)oldValue).DeregisterSeries(this);
                    }

                    if (newValue != null)
                    {
                        ((Axis)newValue).RegisterSeries(this);
                    }

                    if ((XAxis != null && !XAxis.UpdateRange()) ||
                        (newValue == null && oldValue != null))
                    {
                        RenderSeries(false);
                    }
                    break;

                case YAxisPropertyName:
                    if (oldValue != null)
                    {
                        ((Axis)oldValue).DeregisterSeries(this);
                    }

                    if (newValue != null)
                    {
                        ((Axis)newValue).RegisterSeries(this);
                    }
                    if ((YAxis != null && !YAxis.UpdateRange()) ||
                        (newValue == null && oldValue != null))
                    {
                        RenderSeries(false);
                    }
                    break;
            }
        }

        /// <summary>
        /// Calculates range of a given axis based on  X/Y values of data items
        /// </summary>
        /// <param name="axis"></param>
        /// <returns></returns>
        protected override AxisRange GetRange(Axis axis)
        {
            BoxAndWhiskerData myData = this.ItemsSource as BoxAndWhiskerData;
            if (myData == null)
            {
                return base.GetRange(axis);
            }
            if (axis == this.XAxis)
            {
                double min = double.MaxValue;
                double max = double.MinValue;
                foreach (BoxAndWhiskerDataPoint dataPoint in myData)
                {
                    min = System.Math.Min(min, dataPoint.X);
                    max = System.Math.Max(max, dataPoint.X);
                }
                return new AxisRange(min, max);
            }
            else if (axis == this.YAxis)
            {
                double min = double.MaxValue;
                double max = double.MinValue;
                foreach (BoxAndWhiskerDataPoint dataPoint in myData)
                {
                    min = System.Math.Min(min, dataPoint.Y);
                    max = System.Math.Max(max, dataPoint.Y);
                }
                return new AxisRange(min, max);
            }
            else
            {
                return base.GetRange(axis);
            }

        }

        /// <summary>
        /// Gets the item item associated with the specified world position 
        /// when a tooltip must be displayed on the series' marker
        /// </summary>
        protected override object GetItem(Point worldPoint)
        {
            ScalerParams param = new ScalerParams(this.SeriesViewer.ActualWindowRect, this.Viewport, false);

            Point cursorPoint = new Point(this.SeriesViewer.ViewportRect.Width * worldPoint.X,
                                          this.SeriesViewer.ViewportRect.Height * worldPoint.Y);

            BoxAndWhiskerData data = (BoxAndWhiskerData)this.ItemsSource;

            int index = 0;

            foreach (BoxAndWhiskerDataPoint dataPoint in data)
            {
                // scale locations of data point to the series' viewport
                param.IsInverted = this.XAxis.IsInverted;

                double x = this.XAxis.GetScaledValue(index++, param);

                param.IsInverted = this.YAxis.IsInverted;

                double y = this.YAxis.GetScaledValue(dataPoint.Y, param);

                double size = 100 * this.SeriesViewer.ActualWindowRect.Width;
                double left = x - size / 2;
                double top = System.Math.Min(this.YAxis.GetScaledValue(dataPoint.UpperQuartile, param), this.YAxis.GetScaledValue(dataPoint.Maximum, param));
                double bottom = System.Math.Max(this.YAxis.GetScaledValue(dataPoint.LowerQuartile, param), this.YAxis.GetScaledValue(dataPoint.Minimum, param));

                double height = System.Math.Abs(top - bottom);

                Rect itemBounds = new Rect(left, top, size, height);

                if (itemBounds.Contains(cursorPoint))
                {
                    return dataPoint;
                }

                ///////////////////////////////
                //// scale locations of data point to the series' viewport
                //param.IsInverted = this.XAxis.IsInverted;

                ////double x = this.XAxis.GetScaledValue(dataPoint.X, param);
                //double x = this.XAxis.GetScaledValue(index++, param);

                //param.IsInverted = this.YAxis.IsInverted;

                //double y = this.YAxis.GetScaledValue(dataPoint.Y, param);

                //double size = 50;
                //double left = x - size / 2;
                //double top = y - size / 2;

                //Rect itemBounds = new Rect(left, top, size, size);

                //if (itemBounds.Contains(cursorPoint))
                //{
                //    return dataPoint;
                //}
            }

            return null;
        }
    }

}
