﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using System.ComponentModel;
using System.Threading;
using System.Windows;
using System.Windows.Threading;
using System.Windows.Data;
using System.Globalization;

using Infragistics.Controls.Charts;
using System.Windows.Media;

namespace ScatterLineChart
{
    public class DataPoint : ObservableModel
    {
        private double _y;
        public double Y
        {
            get { return _y; }
            set { if (_y == value) return; _y = value; OnPropertyChanged("Y"); }
        }

        private double _x;
        public double X
        {
            get { return _x; }
            set { if (_x == value) return; _x = value; OnPropertyChanged("X"); }
        }

        public new string ToString()
        {
            return String.Format("X {0}, Y {1}", X, Y);
        }

       
        private string _name = string.Empty;
        public string Name
        {
            get
            {
                if (_name != string.Empty)
                    return _name;
                else
                    return this.ToString();
            }
            set { _name = value; }
        }

        public DataPoint( )
        {         
        }

        public DataPoint(double X, double Y)
        {
            _y = Y;
            _x = X;
        }
        public DataPoint(double X, double Y, string Name) : this( X, Y )
        {            
            _name = Name;
        }

    }

    public abstract class ObservableModel : INotifyPropertyChanged, IDataErrorInfo
    {

        #region INotifyPropertyChanged

        public event PropertyChangedEventHandler PropertyChanged;

        protected bool HasPropertyChangedHandler()
        {
            PropertyChangedEventHandler handler = this.PropertyChanged;
            return handler != null;
        }
        protected void OnPropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            PropertyChangedEventHandler handler = this.PropertyChanged;
            if (handler != null)
                handler(sender, e);
        }
        protected void OnPropertyChanged(PropertyChangedEventArgs e)
        {
            OnPropertyChanged(this, e);
        }
        protected void OnPropertyChanged(object sender, string propertyName)
        {
            OnPropertyChanged(sender, new PropertyChangedEventArgs(propertyName));
        }
        protected void OnPropertyChanged(string propertyName)
        {
            OnPropertyChanged(this, new PropertyChangedEventArgs(propertyName));
        }
        protected delegate void PropertyChangedDelegate(object sender, string propertyName);

        //WPF
        //public void OnAsyncPropertyChanged(string propertyName)
        //{
        //    if (PropertyChanged != null)
        //    {
        //        App.Current.Dispatcher.BeginInvoke(DispatcherPriority.Normal,
        //           new ThreadStart(() =>
        //           {
        //              PropertyChanged(this, new PropertyChangedEventArgs(propertyName));
        //         }));
        //    }
        //}
        #endregion
        #region IDataErrorInfo Members
        public ValidationHandler ValidationHandler = new ValidationHandler();
        public string Error
        {
            get { return null; }
        }
        public string this[string columnName]
        {
            get
            {
                if (this.ValidationHandler.BrokenRuleExists(columnName))
                {
                    return this.ValidationHandler[columnName];
                }
                return null;
            }
        }
        #endregion
    }

    public class ValidationHandler
    {
        private Dictionary<string, string> BrokenRules { get; set; }
        public ValidationHandler()
        {
            this.BrokenRules = new Dictionary<string, string>();
        }
        public string this[string property]
        {
            get
            {
                return this.BrokenRules[property];
            }
        }
        public bool BrokenRuleExists(string property)
        {
            return this.BrokenRules.ContainsKey(property);
        }
        public bool ValidateRule(string property, string message, Func<bool> ruleCheck)
        {
            if (!ruleCheck())
            {
                this.BrokenRules.Add(property, message);
                return false;
            }
            RemoveBrokenRule(property);
            return true;
        }
        public void RemoveBrokenRule(string property)
        {
            if (this.BrokenRules.ContainsKey(property))
            {
                this.BrokenRules.Remove(property);
            }
        }
    }


    
}
